from datetime import datetime
from io import BytesIO
import json

from botocore.exceptions import ClientError # noqa

from zfs_uploader import BACKUP_DB_FILE, DATETIME_FORMAT


class BackupDB:
    """ Backup DB object. """

    @property
    def filesystem(self):
        """ ZFS filesystem. """
        return self._filesystem

    def __init__(self, bucket, filesystem):
        """ Create BackupDB object.

        BackupDB is used for storing Backup objects. It does not upload
        backups but serves as a database for backup records.

        Parameters
        ----------
        bucket : Bucket
            S3 Bucket.
        filesystem : str
            ZFS filesystem.

        """
        self._filesystem = filesystem
        self._backups = {}
        self._s3_object = bucket.Object(
            f'{self._filesystem}/{BACKUP_DB_FILE}')

        # initialize from backup.db file if it exists
        self.download()

    def create_backup(self, backup_time, backup_type, s3_key,
                      dependency=None, backup_size=None):
        """ Create backup object and upload `backup.db` file.

        Parameters
        ----------
        backup_time : str
            Backup time in %Y%m%d_%H%M%S format.
        backup_type : str
            Supported backup types are `full` and `inc`.
        s3_key : str
            Backup S3 key.
        dependency : str, optional
            Backup time of dependency in %Y%m%d_%H%M%S format. Used for
            storing the dependent full backup for an incremental backup.
        backup_size : int, optional
            Backup size in bytes.

        """
        if backup_time in self._backups:
            raise ValueError('Backup already exists.')

        if dependency and dependency not in self._backups:
            raise ValueError('Depending on backup does not exist.')

        self._backups.update({
            backup_time: Backup(backup_time, backup_type, self._filesystem,
                                s3_key, dependency, backup_size)
        })

        self.upload()

    def delete_backup(self, backup_time):
        """ Delete backup and upload `backup.db`.

        Parameters
        ----------
        backup_time : str
            Backup time in %Y%m%d_%H%M%S format.

        """
        if _validate_backup_time(backup_time) is False:
            raise ValueError('backup_time is wrong format')

        del self._backups[backup_time]

        self.upload()

    def get_backup(self, backup_time):
        """ Get backup using backup time.

        Parameters
        ----------
        backup_time : str
            Backup time in %Y%m%d_%H%M%S format.

        Returns
        -------
        Backup

        """
        if _validate_backup_time(backup_time) is False:
            raise ValueError('backup_time is wrong format')

        try:
            return self._backups[backup_time]
        except KeyError:
            raise KeyError('Backup does not exist.') from None

    def get_backups(self, backup_type=None):
        """ Get sorted list of backups.

        Parameters
        ----------
        backup_type : str, optional
            Supported backup types are `full` and `inc`.

        Returns
        -------
        list(Backup)
            Sorted list of backups. Most recent backup is last.

        """
        backup_times = sorted(self._backups)

        if backup_type in ['full', 'inc']:
            backups = []
            for time in backup_times:
                backup = self._backups[time]

                if backup.backup_type == backup_type:
                    backups.append(backup)
        elif backup_type is None:
            backups = [self._backups[time] for time in backup_times]
        else:
            raise ValueError('backup_type must be `full` or `inc`')

        return backups

    def get_backup_times(self, backup_type=None):
        """ Get sorted list of backup times.

        Parameters
        ----------
        backup_type : str, optional
            Supported backup types are `full` and `inc`.

        Returns
        -------
        list(str)
            Sorted list of backup times. Most recent backup is last.

        """
        if backup_type in ['full', 'inc']:
            backup_times = []
            for time in sorted(self._backups):
                backup = self._backups[time]

                if backup.backup_type == backup_type:
                    backup_times.append(time)
            return backup_times
        elif backup_type is None:
            return sorted(self._backups)
        else:
            raise ValueError('backup_type must be `full` or `inc`')

    def download(self):
        """ Download backup.db file. """
        try:
            with BytesIO() as f:
                self._s3_object.download_fileobj(f)
                f.seek(0)
                self._backups = json.load(f, object_hook=_json_object_hook)
        except ClientError:
            pass

    def upload(self):
        """ Upload backup.db file. """
        with BytesIO() as f:
            json_str = json.dumps(self._backups, default=_json_default)
            f.write(json_str.encode('utf-8'))
            f.seek(0)
            self._s3_object.upload_fileobj(f)


class Backup:
    """ Backup object. """

    @property
    def backup_time(self):
        """ Backup time. """
        return self._backup_time

    @property
    def backup_type(self):
        """ Backup type. """
        return self._backup_type

    @property
    def filesystem(self):
        """ ZFS filesystem. """
        return self._filesystem

    @property
    def snapshot_name(self):
        """ ZFS snapshot name. """
        return f'{self._filesystem}@{self._backup_time}'

    @property
    def s3_key(self):
        """ S3 key. """
        return self._s3_key

    @property
    def dependency(self):
        """ Backup time of dependency. """
        return self._dependency

    @property
    def backup_size(self):
        """ Backup size in bytes. """
        return self._backup_size

    def __init__(self, backup_time, backup_type, filesystem, s3_key,
                 dependency=None, backup_size=None):
        """ Create Backup object.

        Parameters
        ----------
        backup_time : str
            Backup time in %Y%m%d_%H%M%S format.
        backup_type : str
            Supported backup types are `full` and `inc`.
        filesystem : str
            ZFS filesystem.
        s3_key : str
            Backup S3 key.
        dependency : str, optional
            Backup time of dependency in %Y%m%d_%H%M%S format. Used for
            storing the dependent full backup for an incremental backup.
        backup_size : int, optional
            Backup size in bytes.

        """
        if _validate_backup_time(backup_time):
            self._backup_time = backup_time
        else:
            raise ValueError('backup_time is wrong format')

        if backup_type in ['full', 'inc']:
            self._backup_type = backup_type
        else:
            raise ValueError('backup_type must be `full` or `inc`')

        self._filesystem = filesystem
        self._s3_key = s3_key

        if dependency:
            if not _validate_backup_time(dependency):
                raise ValueError('dependency is wrong format')
        self._dependency = dependency

        self._backup_size = backup_size

    def __eq__(self, other):
        return all((self._backup_time == other._backup_time, # noqa
                    self._backup_type == other._backup_type, # noqa
                    self._filesystem == other._filesystem, # noqa
                    self._s3_key == other._s3_key, # noqa
                    self._dependency == other._dependency, # noqa
                    self._backup_size == other._backup_size # noqa
                    ))

    def __hash__(self):
        return hash((self._backup_time,
                     self._backup_type,
                     self._filesystem,
                     self._s3_key,
                     self._dependency,
                     self._backup_size
                     ))


def _json_default(obj):
    if isinstance(obj, Backup):
        return {
            '_type': 'Backup',
            'backup_time': obj._backup_time, # noqa
            'backup_type': obj._backup_type, # noqa
            'filesystem': obj._filesystem, # noqa
            's3_key': obj._s3_key, # noqa
            'dependency': obj._dependency, # noqa
            'backup_size': obj._backup_size # noqa
        }


def _json_object_hook(dct):
    obj_type = dct.get('_type')
    if obj_type == 'Backup':
        dct_copy = dct.copy()
        del dct_copy['_type']

        return Backup(**dct_copy)
    else:
        return dct


def _validate_backup_time(backup_time):
    try:
        datetime.strptime(backup_time, DATETIME_FORMAT)
    except ValueError:
        return False

    return True



from configparser import ConfigParser
import logging
import os
import sys

from zfs_uploader.job import ZFSjob


class Config:
    """ Wrapper for configuration file. """

    @property
    def jobs(self):
        """ ZFS backup jobs. """
        return self._jobs

    def __init__(self, file_path=None):
        """ Construct Config object from file.

        Parameters
        ----------
        file_path : str
            File path to config file.

        """
        file_path = file_path or 'config.cfg'

        self._logger = logging.getLogger(__name__)
        self._logger.info(f'file_path={file_path} '
                          'msg="Loading configuration file."')

        if not os.path.isfile(file_path):
            self._logger.critical('No configuration file found.')
            sys.exit(1)

        self._cfg = ConfigParser()
        self._cfg.read(file_path)

        default = self._cfg['DEFAULT']
        self._jobs = {}
        for k, v in self._cfg.items():
            if k != 'DEFAULT':
                bucket_name = (v.get('bucket_name') or
                               default.get('bucket_name'))
                access_key = v.get('access_key') or default.get('access_key')
                secret_key = v.get('secret_key') or default.get('secret_key')
                filesystem = k

                if not all((bucket_name, access_key, secret_key)):
                    self._logger.critical(f'file_path={file_path} '
                                          f'filesystem={filesystem}'
                                          'msg="bucket_name, access_key or '
                                          'secret_key is missing from config.')
                    sys.exit(1)

                cron_dict = None
                cron = v.get('cron') or default.get('cron')
                if cron:
                    cron_dict = _create_cron_dict(cron)

                self._jobs[k] = (
                    ZFSjob(
                        bucket_name,
                        access_key,
                        secret_key,
                        filesystem,
                        region=v.get('region') or default.get('region'),
                        cron=cron_dict,
                        max_snapshots=(v.getint('max_snapshots') or
                                       default.getint('max_snapshots')),
                        max_incremental_backups=(
                                v.getint('max_incremental_backups') or
                                default.getint('max_incremental_backups')),
                        storage_class=(v.get('storage_class') or
                                       default.get('storage_class'))
                    )
                )


def _create_cron_dict(cron):
    values = cron.split()

    return {'minute': values[0],
            'hour': values[1],
            'day': values[2],
            'month': values[3],
            'day_of_week': values[4]}



import logging
import os
import sys
from logging.handlers import RotatingFileHandler

import click
from apscheduler.executors.pool import ThreadPoolExecutor
from apscheduler.schedulers.background import BlockingScheduler

from zfs_uploader import __version__
from zfs_uploader.config import Config

LOG_FORMAT = 'time=%(asctime)s.%(msecs)03d level=%(levelname)s %(message)s'


@click.group()
@click.option('--config-path', default='config.cfg',
              help='Config file path.',
              show_default=True)
@click.option('--log-path', default='zfs_uploader.log',
              help='Log file path.',
              show_default=True)
@click.pass_context
def cli(ctx, config_path, log_path):
    logger = logging.getLogger('zfs_uploader')
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(LOG_FORMAT, datefmt='%Y-%m-%dT%H:%M:%S')

    fh = RotatingFileHandler(log_path, maxBytes=5*1024*1024, backupCount=5)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    ch = logging.StreamHandler(sys.stdout)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if not os.path.isfile(config_path):
        print('No configuration file found.')
        sys.exit(1)

    ctx.obj = {
        'config_path': config_path,
        'logger': logger
    }


@cli.command()
@click.pass_context
def backup(ctx):
    """ Start backup job scheduler. """
    config_path = ctx.obj['config_path']
    logger = ctx.obj['logger']

    config = Config(config_path)
    scheduler = BlockingScheduler(
        executors={'default': ThreadPoolExecutor(max_workers=1)}
    )

    for job in config.jobs.values():
        logger.info(f'filesystem={job.filesystem} '
                    f'cron="{job.cron}" '
                    'msg="Adding job."')
        scheduler.add_job(job.start, 'cron', **job.cron,
                          misfire_grace_time=2*60*60,
                          coalesce=True)

    try:
        scheduler.start()
    except (KeyboardInterrupt, SystemExit):
        pass


@cli.command('list')
@click.argument('filesystem', required=False)
@click.pass_context
def list_backups(ctx, filesystem):
    """ List backups. """
    config_path = ctx.obj['config_path']
    logger = ctx.obj['logger']

    logger.setLevel('CRITICAL')

    config = Config(config_path)

    if filesystem:
        job = config.jobs.get(filesystem)
        jobs = {filesystem: job}

        if job is None:
            print('Filesystem does not exist in config file.')
            sys.exit(1)

    else:
        jobs = config.jobs

        if jobs is None:
            print('No filesystems exist in config file.')
            sys.exit(1)

    for filesystem, job in jobs.items():
        print(f'{filesystem}:\n')
        print('{0:<16} {1:<16} {2:<5} {3:<14}'.format('time', 'dependency',
                                                      'type', 'size (bytes)'))
        print('-'*52)
        for b in job.backup_db.get_backups():
            dependency = b.dependency or str(b.dependency)
            backup_size = b.backup_size or str(b.backup_size)
            print(f'{b.backup_time:<16} {dependency:<16} '
                  f'{b.backup_type:<5} {backup_size:<14}')
        print('\n')


@cli.command()
@click.option('--destination', help='Destination filesystem.')
@click.argument('filesystem')
@click.argument('backup-time', required=False)
@click.pass_context
def restore(ctx, destination, filesystem, backup_time):
    """ Restore from backup.

    Defaults to most recent backup if backup-time is not specified.

    WARNING: If restoring to a file system that already exists, snapshots
    and data that were written after the backup will be destroyed. Set
    `destination` in order to restore to a new file system.

    """
    config_path = ctx.obj['config_path']

    config = Config(config_path)
    job = config.jobs.get(filesystem)

    if job is None:
        print('Filesystem does not exist.')
        sys.exit(1)

    job.restore(backup_time, destination)

    print('Restore successful.')


@cli.command(help='Print version.')
def version():
    print(__version__)


if __name__ == '__main__':
    cli()
